(* 多相型関数をインスタンス化する *)

open Type
open Id
open FlatSyntax

module M3 = MyMap.Make(struct
  type t = Type.t list
  let compare = Pervasives.compare
end)

(* 型 *)
let typeof x =
  match x with
  | Unit -> Tunit
  | Int _ | Bool _ -> Tint
  | Float _ -> Tfloat
  | Var x -> match x.typ with
    | Tvar{ contents = None } -> failwith
          ("cannot determine type of " ^ Id.to_string x)
    | t -> t

(* 型変数と代入される型の対応を作る *)
let make_tenv formal_arg_ts actual_arg_ts =
  let rec aux acc t1 t2 =
    match t1,t2 with
    | Tvar({ contents = None } as v), t ->
        if List.mem_assq v acc then acc else (v,t) :: acc
    | Tarray t1, Tarray t2 -> aux acc t1 t2
    | Ttuple t1s, Ttuple t2s -> List.fold_left2 aux acc t1s t2s
    | Tfun(t1s,t1), Tfun(t2s,t2) -> List.fold_left2 aux (aux acc t1 t2) t1s t2s
    | _ -> assert (t1 == t2); acc
  in
  List.fold_left2 aux [] formal_arg_ts actual_arg_ts

(* 型変数に代入した新しい型を作る *)
let rec copy_t tenv t =
  match t with
  | Tvar({ contents = None } as v) -> List.assq v tenv
  | Tarray t -> Tarray(copy_t tenv t)
  | Ttuple ts -> Ttuple(List.map (copy_t tenv) ts)
  | Tfun(ts,t) -> Tfun(List.map (copy_t tenv) ts, copy_t tenv t)
  | Tvar _ | ForAll _ -> assert false
  | _ -> t

(* 型変数に代入した新しい式を作る *)
let rec copy env tenv e =
  let find x = match x with
  | Var y -> (try Var(M.find y env) with Not_found -> x)
  | _ -> x
  in
  let copy_id x = new_id_t x.name (copy_t tenv x.typ) in
  match e with
  | Let(xs,e1,e2) ->
      let xs_ = List.map copy_id xs in
      Let(xs_, copy env tenv e1, copy (M.add_list2 xs xs_ env) tenv e2)
  | LetRec(x,ys,e1,e2) ->
      let x_ = copy_id x in
      let ys_ = List.map copy_id ys in
      let env2 = M.add x x_ env in
      let env1 = M.add_list2 ys ys_ env2 in
      LetRec(x_, ys_, copy env1 tenv e1, copy env2 tenv e2)
  | If(x,e1,e2) -> If(find x, copy env tenv e1, copy env tenv e2)
  | Op(x,ys) -> Op(find x, List.map find ys)
  | Tuple xs -> Tuple(List.map find xs)
  | V x -> V(find x)
  | ExtApp(x,ys) -> ExtApp(x, List.map find ys)

(* 多相型関数 -> 型 -> 型を限定した関数 *)
let instantiate (x,ys,e) ts =
  match x.typ with
  | ForAll(vs, Tfun(ts_, t)) ->
      let tenv = make_tenv ts_ ts in
      let tx = Tfun(ts, copy_t tenv t) in
      let x_ = new_id_t x.name tx in
      let ys_ = List.map2 (fun y t -> new_id_t y.name t) ys ts in
      let env = M.add_list2 ys ys_ (M.add x x_ M.empty) in
      let e_ = copy env tenv e in
      (x_, ys_, e_)
  | _ -> failwith "Expansion.instantiate"

(* 多相型関数と型から、すでに生成された関数への写像 *)
let env = ref M.empty

(* 使用する箇所を置き換えたり LetRec を追加したり *)
let rec g funs e =
  let find x ts env = M3.find ts (M.find x env) in
  let add x ts y env = M.add x
    (M3.add ts y (try M.find x env with Not_found -> M3.empty))
    env
  in
  match e with
  | Op(_, xs)
    when List.exists (function Var x -> M.mem x funs | _ -> false) xs ->
      (* この辺はexpansionじゃ無理ぽ *)
      failwith "polymorphic closure not supported"
  | V _ | Tuple _ | ExtApp _ -> e
  | Op(Var x, ys) when not (M.mem x funs) -> e
  | Op(Var x, ys) -> begin try
      let ts = List.map typeof ys in
      let (z,_,_) = try find x ts !env with Not_found ->
        let name,args,body = instantiate (M.find x funs) ts in
        env := add x ts (name,[],V Unit) !env;
        let name,args,body as f = name,args, g funs body in
        env := add x ts f !env; f
      in
      Op(Var z, ys) with _ -> e end
  | Op _ -> assert false
  | If(x,e1,e2) -> If(x, g funs e1, g funs e2)
  | Let(xs,e1,e2) -> Let(xs, g funs e1, g funs e2)
  | LetRec(x,ys,e1,e2) ->
      let e_ = match x.typ with
        | ForAll _ -> LetRec(x,ys,e1, g (M.add x (x,ys,e1) funs) e2)
        | _ -> LetRec(x,ys, g funs e1, g funs e2)
      in
      try M3.fold
        (fun _ (x,ys,e1) e2 -> LetRec(x,ys,e1,e2))
        (M.find x !env) e_
      with Not_found -> e_

(* エントリーポイント *)
let f e =
  env := M.empty;
  let ret = g M.empty e in
  env := M.empty; (* メモリ節約 :-p *)
  ret
